@@ -151,6 +151,19 @@ module TestRubyParserShared
     assert_parse rb, pt
   end
 
+  def test_attrasgn_array_arg
+    rb = "a[[1, 2]] = 3"
+    pt = s(:attrasgn,
+           s(:call, nil, :a),
+           :[]=,
+           s(:array,
+             s(:lit, 1),
+             s(:lit, 2)),
+           s(:lit, 3))
+
+    assert_parse rb, pt
+  end
+
   def test_block_append
     head = s(:args)
     tail = s(:zsuper)
@@ -281,6 +294,13 @@ module TestRubyParserShared
     assert_parse rb, pt
   end
 
+  def test_begin_else_return_value
+    rb = "begin; else 2; end"
+    pt = s(:lit, 2)
+
+    assert_parse rb, pt
+  end
+
   def test_bug_comment_eq_begin
     rb = "\n\n#\n=begin\nblah\n=end\n\n"
     pt = nil
@@ -290,6 +310,13 @@ module TestRubyParserShared
     assert_equal exp, processor.lexer.comments
   end
 
+  def test_eq_begin_why_wont_people_use_their_spacebar?
+    rb = "h[k]=begin\n       42\n     end"
+    pt = s(:attrasgn, s(:call, nil, :h), :[]=, s(:call, nil, :k), s(:lit, 42))
+
+    assert_parse rb, pt
+  end
+
   def test_bug_call_arglist_parens
     rb = 'g ( 1), 2'
     pt = s(:call, nil, :g, s(:lit, 1), s(:lit, 2))
@@ -360,6 +387,14 @@ module TestRubyParserShared
     assert_equal expected_env, processor.env.all
   end
 
+  def test_lasgn_ivar_env
+    rb = '@a = 42'
+    pt = s(:iasgn, :@a, s(:lit, 42))
+
+    assert_parse rb, pt
+    assert_empty processor.env.all
+  end
+
   def test_list_append
     a = s(:lit, 1)
     b = s(:lit, 2)
@@ -539,6 +574,13 @@ module TestRubyParserShared
     assert_parse rb, pt
   end
 
+  def test_str_pct_q
+    rb = "%q{a b c}"
+    pt = s(:str, "a b c")
+
+    assert_parse rb, pt
+  end
+
   # def test_str_pct_nested_nested
   #   rb = "%{ { #\{ \"#\{1}\" } } }"
   #   assert_equal " { 1 } ", eval(rb)
@@ -561,14 +603,8 @@ module TestRubyParserShared
     assert_parse rb, pt
   end
 
-  STARTING_LINE = {
-    "case_no_expr"                       => 2, # TODO this should be 1
-    "structure_unused_literal_wwtt"      => 3, # yes, 3... odd test
-  }
-
   def after_process_hook klass, node, data, input_name, output_name
-    expected = STARTING_LINE[node] || 1
-    assert_equal expected, @result.line, "should have proper line number"
+    assert_equal 1, @result.line, "should have proper line number"
   end
 
   def test_parse_line_block
@@ -590,6 +626,36 @@ module TestRubyParserShared
     assert_same result.file, result.call.file
   end
 
+  def test_parse_line_block_inline_comment
+    rb = "a\nb # comment\nc"
+    pt = s(:block,
+           s(:call, nil, :a).line(1),
+           s(:call, nil, :b).line(2),
+           s(:call, nil, :c).line(3))
+
+    assert_parse rb, pt
+  end
+
+  def test_parse_line_block_inline_multiline_comment
+    rb = "a\nb # comment\n# another comment\nc"
+    pt = s(:block,
+           s(:call, nil, :a).line(1),
+           s(:call, nil, :b).line(2),
+           s(:call, nil, :c).line(4)).line(1)
+
+    assert_parse rb, pt
+  end
+
+  def test_parse_line_block_inline_comment_leading_newlines
+    rb = "\n\n\na\nb # comment\n# another comment\nc"
+    pt = s(:block,
+           s(:call, nil, :a).line(4),
+           s(:call, nil, :b).line(5),
+           s(:call, nil, :c).line(7)).line(4)
+
+    assert_parse rb, pt
+  end
+
   def test_parse_line_call_no_args
     rb = "f do |x, y|\n  x + y\nend"
 
@@ -604,12 +670,28 @@ module TestRubyParserShared
     assert_equal 2, result[3].line,   "call should have line number"
   end
 
+  def test_parse_line_call_ivar_line_break_paren
+    rb = "a(@b\n)"
+    pt = s(:call, nil, :a, s(:ivar, :@b).line(1)).line(1)
+
+    assert_parse rb, pt
+  end
+
+  def test_parse_line_call_ivar_arg_no_parens_line_break
+    rb = "a @b\n"
+    pt = s(:call, nil, :a, s(:ivar, :@b).line(1)).line(1)
+
+    assert_parse rb, pt
+  end
+
   def test_parse_line_defn_no_parens
     pt = s(:defn, :f, s(:args), s(:nil))
 
     rb = "def f\nend"
     assert_parse_line rb, pt, 1
 
+    processor.reset
+
     rb = "def f\n\nend"
     assert_parse_line rb, pt, 1
   end
@@ -629,6 +711,53 @@ module TestRubyParserShared
     assert_equal 4, body.return.line, "return should have line number"
   end
 
+  def test_parse_line_evstr_after_break
+    rb = "\"a\"\\\n\"\#{b}\""
+    pt = s(:dstr, "a",
+           s(:evstr,
+             s(:call, nil, :b).line(2)).line(2)).line(1)
+
+    assert_parse rb, pt
+  end
+
+  def test_parse_line_dstr_newline
+    skip "dstr line numbers are just gonna be screwed for a while..."
+
+    rb = <<-'CODE'
+            "a\n#{
+            }"
+            true
+    CODE
+
+    pt = s(:block,
+           s(:dstr, "a\n",
+             s(:evstr)).line(1),
+           s(:true).line(3))
+
+    assert_parse rb, pt
+  end
+
+  def test_parse_line_heredoc_evstr
+    skip "heredoc line numbers are just gonna be screwed for a while..."
+
+    rb = "<<-A\na\n\#{b}\nA"
+    pt = s(:dstr, "a\n",
+           s(:evstr,
+             s(:call, nil, :b).line(3)),
+             s(:str, "\n")).line(1)
+
+    assert_parse rb, pt
+  end
+
+  def test_parse_line_multiline_str
+    rb = "\"a\nb\"\n1"
+    pt = s(:block,
+           s(:str, "a\nb").line(1),
+           s(:lit, 1).line(3)).line(1)
+
+    assert_parse rb, pt
+  end
+
   def test_parse_line_iter_call_parens
     rb = "f(a) do |x, y|\n  x + y\nend"
 
@@ -648,15 +777,12 @@ module TestRubyParserShared
     rb = "f a do |x, y|\n  x + y\nend"
 
     pt = s(:iter,
-           s(:call, nil, :f, s(:call, nil, :a)),
-           s(:args, :x, :y),
-           s(:call, s(:lvar, :x), :+, s(:lvar, :y)))
+           s(:call, nil, :f, s(:call, nil, :a).line(1)).line(1),
+           s(:args, :x, :y).line(1),
+           s(:call, s(:lvar, :x).line(2), :+,
+             s(:lvar, :y).line(2)).line(2)).line(1)
 
-    assert_parse_line rb, pt, 1
-
-    assert_equal 1, result[1].line,   "call should have line number"
-    assert_equal 1, result[2].line,   "masgn should have line number"
-    assert_equal 2, result[3].line,   "call should have line number"
+    assert_parse rb, pt
   end
 
   def test_parse_line_heredoc
@@ -718,6 +844,24 @@ module TestRubyParserShared
     assert_equal 3, result.if.return.lit.line
   end
 
+  def test_parse_line_str_with_newline_escape
+    rb = 'a("\n", true)'
+    pt = s(:call, nil, :a,
+           s(:str, "\n").line(1),
+           s(:true).line(1))
+
+    assert_parse rb, pt
+  end
+
+  def test_parse_line_trailing_newlines
+    rb = "a \nb"
+    pt = s(:block,
+           s(:call, nil, :a).line(1),
+           s(:call, nil, :b).line(2)).line(1)
+
+    assert_parse rb, pt
+  end
+
   def test_bug_and
     rb = "true and []"
     pt = s(:and, s(:true), s(:array))
@@ -777,13 +921,15 @@ module TestRubyParserShared
     Ruby20Parser === self.processor
   end
 
+  def ruby21
+    Ruby21Parser === self.processor
+  end
+
   def test_bug_comma
     val = if ruby18 then
             s(:lit, 100)
-          elsif ruby19  or ruby20 then
-            s(:str, "d")
           else
-            raise "wtf"
+            s(:str, "d")
           end
 
     rb = "if test ?d, dir then end"
@@ -834,7 +980,7 @@ module TestRubyParserShared
     rb = "not(a)"
     pt = if ruby18 then
            s(:not, s(:call, nil, :a))
-         elsif ruby19 or ruby20 then
+         elsif ruby19 or ruby20 or ruby21 then
            s(:call, s(:call, nil, :a), :"!")
          else
            raise "wtf"
@@ -965,7 +1111,7 @@ module TestRubyParserShared
                s(:call, nil, :p, s(:str, "a").line(2)).line(2),
                s(:lasgn, :b, s(:lit, 1).line(3)).line(3),
                s(:call, nil, :p, s(:lvar, :b).line(4)).line(4),
-               s(:lasgn, :c, s(:lit, 1).line(5)).line(5)).line(2), # TODO line 2?
+               s(:lasgn, :c, s(:lit, 1).line(5)).line(5)).line(2),
              nil).line(1),
            s(:call, nil, :a).line(7)).line(1)
 
@@ -1236,15 +1382,6 @@ module TestRubyParserShared
     assert_parse rb, pt
   end
 
-  def test_aref_args_lit_assocs
-    skip if ruby18
-
-    rb = "[1, 2 => 3]"
-    pt = s(:array, s(:lit, 1), s(:hash, s(:lit, 2), s(:lit, 3)))
-
-    assert_parse rb, pt
-  end
-
   def test_BEGIN
     rb = "BEGIN { 42 }"
     pt = s(:iter, s(:preexe), s(:args), s(:lit, 42))
@@ -1305,15 +1442,6 @@ module TestRubyParserShared
     assert_parse rb, pt
   end
 
-  def test_block_decomp_arg_splat
-    skip "not that smart yet" if ruby18 # HACK
-
-    rb = "a { |(b, *)| }"
-    pt = s(:iter, s(:call, nil, :a), s(:args, s(:masgn, :b, :*)))
-
-    assert_parse rb, pt
-  end
-
   def test_masgn_arg_ident
     rb = "a, b.C = d"
     pt = s(:masgn,
@@ -1413,7 +1541,6 @@ module TestRubyParserShared
   end
 
   def test___ENCODING__
-    # skip "lexer bug" if ruby18
     rb = "__ENCODING__"
     pt = if Ruby18Parser === processor then
            s(:call, nil, :__ENCODING__)
@@ -1447,7 +1574,21 @@ module TestRubyParserShared
   end
 end
 
-module TestRubyParserShared1920
+module TestRubyParserShared19to21
+  def test_aref_args_lit_assocs
+    rb = "[1, 2 => 3]"
+    pt = s(:array, s(:lit, 1), s(:hash, s(:lit, 2), s(:lit, 3)))
+
+    assert_parse rb, pt
+  end
+
+  def test_block_decomp_arg_splat
+    rb = "a { |(b, *)| }"
+    pt = s(:iter, s(:call, nil, :a), s(:args, s(:masgn, :b, :*)))
+
+    assert_parse rb, pt
+  end
+
   def test_block_call_operation_dot
     rb = "a.b c do end.d"
     pt = s(:call,
@@ -1919,9 +2060,8 @@ module TestRubyParserShared1920
   end
 
   def test_mlhs_keyword
-    skip "Breaks on 1.9 and 2.0 parser but valid" # HACK
     rb = "a.!=(true, true)"
-    pt = 42
+    pt = s(:call, s(:call, nil, :a), :"!=", s(:true), s(:true))
 
     assert_parse rb, pt
   end
@@ -1934,6 +2074,70 @@ module TestRubyParserShared1920
   end
 end
 
+module TestRubyParserShared20to21
+  def test_defn_kwarg_kwsplat
+    rb = "def a(b: 1, **c) end"
+    pt = s(:defn, :a, s(:args, s(:kwarg, :b, s(:lit, 1)), :"**c"), s(:nil))
+
+    assert_parse rb, pt
+  end
+
+  def test_defn_kwarg_env
+    rb = "def test(**testing) test_splat(**testing) end"
+    pt = s(:defn, :test, s(:args, :"**testing"),
+           s(:call, nil, :test_splat, s(:kwsplat, s(:lvar, :testing))))
+
+    assert_parse rb, pt
+  end
+
+  def test_call_arg_kwsplat
+    rb = "a(b, **1)"
+    pt = s(:call, nil, :a, s(:call, nil, :b), s(:kwsplat, s(:lit, 1)))
+
+    assert_parse rb, pt
+  end
+
+  def test_call_kwsplat
+    rb = "a(**1)"
+    pt = s(:call, nil, :a, s(:kwsplat, s(:lit, 1)))
+
+    assert_parse rb, pt
+  end
+
+  def test_iter_kwarg
+    rb = "a { |b: 1| }"
+    pt = s(:iter, s(:call, nil, :a), s(:args, s(:kwarg, :b, s(:lit, 1))))
+
+    assert_parse rb, pt
+  end
+
+  def test_iter_kwarg_kwsplat
+    rb = "a { |b: 1, **c| }"
+    pt = s(:iter, s(:call, nil, :a), s(:args, s(:kwarg, :b, s(:lit, 1)), :"**c"))
+
+    assert_parse rb, pt
+  end
+
+  def test_block_kwarg_lvar
+    rb = "bl { |kw: :val| kw }"
+    pt = s(:iter, s(:call, nil, :bl), s(:args, s(:kwarg, :kw, s(:lit, :val))),
+           s(:lvar, :kw))
+
+    assert_parse rb, pt
+  end
+
+  def test_block_kwarg_lvar_multiple
+    rb = "bl { |kw: :val, kw2: :val2 | kw }"
+    pt = s(:iter, s(:call, nil, :bl),
+           s(:args,
+             s(:kwarg, :kw, s(:lit, :val)),
+             s(:kwarg, :kw2, s(:lit, :val2))),
+           s(:lvar, :kw))
+
+    assert_parse rb, pt
+  end
+end
+
 class TestRubyParser < Minitest::Test
   def test_parse
     processor = RubyParser.new
@@ -1973,6 +2177,16 @@ class TestRuby18Parser < RubyParserTestCase
     self.processor = Ruby18Parser.new
   end
 
+  def test_call_space_before_paren_args
+    rb = "a (:b, :c, :d => :e)"
+    pt = s(:call, nil, :a,
+           s(:lit, :b),
+           s(:lit, :c),
+           s(:hash, s(:lit, :d), s(:lit, :e)))
+
+    assert_parse rb, pt
+  end
+
   def test_flip2_env_lvar
     rb = "if a..b then end"
     pt = s(:if, s(:flip2, s(:call, nil, :a), s(:call, nil, :b)), nil, nil)
@@ -2162,7 +2376,7 @@ end
 
 class TestRuby19Parser < RubyParserTestCase
   include TestRubyParserShared
-  include TestRubyParserShared1920
+  include TestRubyParserShared19to21
 
   def setup
     super
@@ -2762,7 +2976,8 @@ end
 
 class TestRuby20Parser < RubyParserTestCase
   include TestRubyParserShared
-  include TestRubyParserShared1920
+  include TestRubyParserShared20to21
+  include TestRubyParserShared19to21
 
   def setup
     super
@@ -2770,6 +2985,38 @@ class TestRuby20Parser < RubyParserTestCase
     self.processor = Ruby20Parser.new
   end
 
+  def test_block_call_dot_op2_brace_block
+    rb = "a.b c() do d end.e do |f| g end"
+    pt = s(:iter,
+           s(:call,
+             s(:iter,
+               s(:call, s(:call, nil, :a), :b, s(:call, nil, :c)),
+               s(:args),
+               s(:call, nil, :d)),
+             :e),
+           s(:args, :f),
+           s(:call, nil, :g))
+
+
+    assert_parse rb, pt
+  end
+
+  def test_block_call_dot_op2_cmd_args_do_block
+    rb = "a.b c() do d end.e f do |g| h end"
+    pt = s(:iter,
+           s(:call,
+             s(:iter,
+               s(:call, s(:call, nil, :a), :b, s(:call, nil, :c)),
+               s(:args),
+               s(:call, nil, :d)),
+             :e,
+             s(:call, nil, :f)),
+           s(:args, :g),
+           s(:call, nil, :h))
+
+    assert_parse rb, pt
+  end
+
   def test_defn_kwarg_val
     rb = "def f(a, b:1) end"
     pt = s(:defn, :f, s(:args, :a, s(:kwarg, :b, s(:lit, 1))), s(:nil))
@@ -2794,6 +3041,20 @@ class TestRuby20Parser < RubyParserTestCase
     assert_parse rb, pt
   end
 
+  def test_defn_kwarg_lvar
+    rb = "def fun(kw: :val); kw; end"
+    pt = s(:defn, :fun, s(:args, s(:kwarg, :kw, s(:lit, :val))), s(:lvar, :kw))
+
+    assert_parse rb, pt
+  end
+
+  def test_defn_kwarg_no_parens
+    rb = "def f a: 1\nend"
+    pt = s(:defn, :f, s(:args, s(:kwarg, :a, s(:lit, 1))), s(:nil))
+
+    assert_parse rb, pt
+  end
+
   def test_defn_powarg
     rb = "def f(**opts) end"
     pt = s(:defn, :f, s(:args, :"**opts"), s(:nil))
@@ -2867,49 +3128,13 @@ class TestRuby20Parser < RubyParserTestCase
     assert_parse rb, pt
   end
 
-  def test_defn_unary_not # TODO: this needs to work on 1.9
-    skip "Not yet"
+  def test_defn_unary_not
     rb = "def !@; true; end" # I seriously HATE this
     pt = s(:defn, :"!@", s(:args), s(:true))
 
     assert_parse rb, pt
   end
 
-  def test_defn_kwarg_kwsplat
-    rb = "def a(b: 1, **c) end"
-    pt = s(:defn, :a, s(:args, s(:kwarg, :b, s(:lit, 1)), :"**c"), s(:nil))
-
-    assert_parse rb, pt
-  end
-
-  def test_call_arg_kwsplat
-    rb = "a(b, **1)"
-    pt = s(:call, nil, :a, s(:call, nil, :b), s(:kwsplat, s(:lit, 1)))
-
-    assert_parse rb, pt
-  end
-
-  def test_call_kwsplat
-    rb = "a(**1)"
-    pt = s(:call, nil, :a, s(:kwsplat, s(:lit, 1)))
-
-    assert_parse rb, pt
-  end
-
-  def test_iter_kwarg
-    rb = "a { |b: 1| }"
-    pt = s(:iter, s(:call, nil, :a), s(:args, s(:kwarg, :b, s(:lit, 1))))
-
-    assert_parse rb, pt
-  end
-
-  def test_iter_kwarg_kwsplat
-    rb = "a { |b: 1, **c| }"
-    pt = s(:iter, s(:call, nil, :a), s(:args, s(:kwarg, :b, s(:lit, 1)), :"**c"))
-
-    assert_parse rb, pt
-  end
-
   def test_iter_array_curly
     rb = "f :a, [:b] { |c, d| }" # yes, this is bad code... that's their problem
     pt = s(:iter,
@@ -2919,3 +3144,43 @@ class TestRuby20Parser < RubyParserTestCase
     assert_parse rb, pt
   end
 end
+
+class TestRuby21Parser < RubyParserTestCase
+  include TestRubyParserShared
+  include TestRubyParserShared19to21
+  include TestRubyParserShared20to21
+
+  def setup
+    super
+
+    self.processor = Ruby21Parser.new
+  end
+
+  def test_f_kw
+    rb = "def x k:42; end"
+    pt = s(:defn, :x, s(:args, s(:kwarg, :k, s(:lit, 42))), s(:nil))
+
+    assert_parse rb, pt
+  end
+
+  def test_f_kw__required
+    rb = "def x k:; end"
+    pt = s(:defn, :x, s(:args, s(:kwarg, :k)), s(:nil))
+
+    assert_parse rb, pt
+  end
+
+  def test_block_kw
+    rb = "-> (k:42) { }"
+    pt = s(:iter, s(:call, nil, :lambda), s(:args, s(:kwarg, :k, s(:lit, 42))))
+
+    assert_parse rb, pt
+  end
+
+  def test_block_kw__required
+    rb = "-> (k:) { }"
+    pt = s(:iter, s(:call, nil, :lambda), s(:args, s(:kwarg, :k)))
+
+    assert_parse rb, pt
+  end
+end
